@@ -10,9 +10,21 @@ PyBambooHR.py contains a class by the same name with functions that correspond
 to BambooHR API calls defined at http://www.bamboohr.com/api/documentation/.
 """
 
+import datetime
 import requests
+import utils
+
+# Python 3 basestring compatibility:
+try:
+    unicode = unicode
+except NameError:
+    # unicode is undefined: We are running Python 3
+    unicode = str
+    basestring = (str, bytes)
+else:
+    # unicode is defined: We are running Python 2
+    bytes = str
 
-from utils import camelcase_keys, underscore_keys, underscore_to_camelcase
 
 class PyBambooHR(object):
     """
@@ -170,6 +182,20 @@ class PyBambooHR(object):
         xml = "<employee>\n{}</employee>".format(xml_fields)
         return xml
 
+    def _format_row_xml(self, row):
+        """
+        Utility method for turning an row dictionary into valid xml for
+        entering or updating a row into the table
+
+        @param employee: Dictionary containing row data information.
+        """
+        xml_fields = ''
+        for k, v in row.iteritems():
+            xml_fields += '\t<field id="{0}">{1}</field>\n'.format(k, v)
+
+        xml = "<row>\n{}</row>".format(xml_fields)
+        return xml
+
     def _format_report_xml(self, fields, title='My Custom Report', report_format='pdf'):
         """
         Utility method for turning an employee dictionary into valid employee xml.
@@ -192,7 +218,7 @@ class PyBambooHR(object):
         @param employee: Dictionary containing employee information.
         @return: Dictionary contianing new employee URL and ID.
         """
-        employee = camelcase_keys(employee)
+        employee = utils.camelcase_keys(employee)
         if not employee.get('firstName') or not employee.get('lastName'):
             raise UserWarning("The 'firstName' and 'lastName' keys are required.")
 
@@ -212,7 +238,7 @@ class PyBambooHR(object):
         @param employee: Dictionary containing employee information.
         @return: Boolean of request success (Status Code == 200).
         """
-        employee = camelcase_keys(employee)
+        employee = utils.camelcase_keys(employee)
         xml = self._format_employee_xml(employee)
         url = self.base_url + 'employees/{0}'.format(id)
         r = requests.post(url, data=xml, headers=self.headers, auth=(self.api_key, ''))
@@ -234,7 +260,7 @@ class PyBambooHR(object):
         data = r.json()
         employees = data['employees']
         if self.underscore_keys:
-            employees = [underscore_keys(employee) for employee in employees]
+            employees = [utils.underscore_keys(employee) for employee in employees]
 
         return employees
 
@@ -249,7 +275,7 @@ class PyBambooHR(object):
         """
         get_fields = []
 
-        field_list = [underscore_to_camelcase(field) for field in field_list] if field_list else None
+        field_list = [utils.underscore_to_camelcase(field) for field in field_list] if field_list else None
 
         if field_list:
             for f in field_list:
@@ -272,10 +298,47 @@ class PyBambooHR(object):
         employee = r.json()
 
         if self.underscore_keys:
-            employee = underscore_keys(employee)
+            employee = utils.underscore_keys(employee)
 
         return employee
 
+    def add_row(self, table_name, employee_id, row):
+        """
+        API method for adding a row to a table
+        http://www.bamboohr.com/api/documentation/tables.php
+
+        @param table_name: string of table's name
+        @param employee_id: string of employee id
+        @param xml: string of xml data to send to BambooHR
+        """
+        row = utils.camelcase_keys(row)
+        xml = self._format_row_xml(row)
+        url = self.base_url + \
+            "employees/{0}/tables/{1}/".format(employee_id, table_name)
+        r = requests.post(url, data=xml, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+
+        return True
+
+    def update_row(self, table_name, employee_id, row_id, row):
+        """
+        API method for updating a row in a table
+        http://www.bamboohr.com/api/documentation/tables.php
+
+        @param table_name: string of table's name
+        @param employee_id: string of employee id
+        @param row_id: string of id of row in table to update
+        @param xml: string of xml data to send to BambooHR
+        """
+        row = utils.camelcase_keys(row)
+        xml = self._format_row_xml(row)
+        url = self.base_url + \
+            "employees/{0}/tables/{1}/{2}/".format(employee_id, table_name, row_id)
+        r = requests.post(url, data=xml, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+
+        return True
+
     def request_company_report(self, report_id, report_format='json', output_filename=None, filter_duplicates=True):
         """
         API method for returning a company report by report ID.
@@ -337,7 +400,7 @@ class PyBambooHR(object):
             raise UserWarning("You requested an invalid report type. Valid values are: {0}".format(','.join([k for k in report_formats])))
 
         get_fields = []
-        field_list = [underscore_to_camelcase(field) for field in field_list] if field_list else None
+        field_list = [utils.underscore_to_camelcase(field) for field in field_list] if field_list else None
         if field_list:
             for f in field_list:
                 if not self.employee_fields.get(f):
@@ -371,3 +434,49 @@ class PyBambooHR(object):
                     handle.write(block)
 
         return result
+
+    def get_tabular_data(self, table_name, employee_id='all'):
+        """
+        API method to retrieve tabular data for an employee, or all employees if employee_id argument is 'all' (the default).
+        See http://www.bamboohr.com/api/documentation/tables.php for a list of available tables.
+
+        @return A dictionary with employee ID as key and a list of dictionaries, each dictionary showing
+        the values of the table's fields for a particular date, which is stored by key 'date' in the dictionary.
+        """
+        url = self.base_url + 'employees/{}/tables/{}'.format(employee_id, table_name)
+        r = requests.get(url, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+
+        return utils.transform_tabular_data(r.content)
+
+    def get_employee_changes(self, since=None):
+        """
+        Returns a list of dictionaries, each with id, action, and lastChanged keys, representing
+        the employee records that have changed since the datetime object passed in the since= argument.
+
+        @return List of dictionaries, each with id, action, and lastChanged keys.
+        """
+        if not isinstance(since, datetime.datetime):
+            raise ValueError("Error: since argument must be a datetime.datetime instance")
+
+        url = self.base_url + 'employees/changed/'
+        params = {'since': since.strftime('%Y-%m-%dT%H:%M:%SZ')}
+        r = requests.get(url, params=params, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+
+        return utils.transform_change_list(r.content)
+
+    def get_whos_out(self, start_date=None, end_date=None):
+        start_date = utils.resolve_date_argument(start_date)
+        end_date = utils.resolve_date_argument(end_date)
+
+        url = self.base_url + 'time_off/whos_out'
+        params = {}
+        if start_date:
+            params['start'] = start_date
+        if end_date:
+            params['end'] = end_date
+        r = requests.get(url, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+
+        return utils.transform_whos_out(r.content)
